﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using Moq;
using System.Collections.Generic;
using System.IO.Abstractions.TestingHelpers;
using TraceLabSDK;
using TraceLabSDK.Component.Config;
using TraceLabSDK.Types;

namespace SentenceParser.Tests
{
    [TestClass]
    public class SentenceParserComponentTests
    {
        private MockFileSystem fileSystem;

        private string basePath;

        private IDictionary<string, object> workspace;

        private SentenceParserComponent component;

        [TestInitialize]
        public void TestInitialize()
        {
            this.fileSystem = new MockFileSystem();
            this.workspace = new Dictionary<string, object>();

            var logger = new Mock<ComponentLogger>();
            this.component = new SentenceParserComponent(logger.Object, this.fileSystem);

            this.basePath = this.fileSystem.Directory.GetCurrentDirectory();
            this.component.Configuration.Artifacts = new FilePath();
            this.component.Configuration.Artifacts.Init(this.fileSystem.Path.Combine(this.basePath, "Artifacts.txt"));
            this.component.Configuration.OutputDirectory = new DirectoryPath();
            this.component.Configuration.OutputDirectory.Init(this.fileSystem.Path.Combine(this.basePath, "Sentences"));

            var workspace = new Mock<IWorkspace>();
            workspace.Setup(mock => mock.Load(It.IsAny<string>()))
                .Returns<string>(str => this.workspace[str]);
            workspace.Setup(mock => mock.Store(It.IsAny<string>(), It.IsAny<object>()))
                .Callback<string, object>((str, obj) => this.workspace[str] = obj);
            this.component.Workspace = workspace.Object;
        }

        [TestMethod]
        public void TestConfigurationIsCustomType()
        {
            Assert.IsInstanceOfType(this.component.Configuration, typeof(SentenceParserConfiguration));
        }

        [TestMethod]
        public void TestComputeThrowsComponentExceptionWhenArtifactsPathIsInvalid()
        {
            Assert.ThrowsException<ComponentException>(() => this.component.Compute());
        }

        [TestMethod]
        public void TestComputeThrowsComponentExceptionWhenOutputDirectoryPathIsInvalid()
        {
            Assert.ThrowsException<ComponentException>(() => this.component.Compute());
        }

        [TestMethod]
        public void TestComputeParsesSentences()
        {
            var sentences = new MockFileData("This is the first sentence. Now this is the second.");
            this.fileSystem.AddFile(this.fileSystem.Path.Combine(this.basePath, "Artifacts.txt"), sentences);
            this.fileSystem.AddDirectory(this.fileSystem.Path.Combine(this.basePath, "Sentences"));

            this.component.Compute();
            var result = this.component.Workspace.Load("outputName") as TLArtifactsCollection;

            var expected = new TLArtifactsCollection();
            expected.Add(new TLArtifact("1", "This is the first sentence."));
            expected.Add(new TLArtifact("2", "Now this is the second."));
            CollectionAssert.AreEquivalent(expected, result);
        }
    }
}
