﻿using System.Collections.Generic;
using System.ComponentModel;
using System.IO.Abstractions;
using System.Linq;
using TraceLabSDK;
using TraceLabSDK.Types;

namespace SentenceParser
{
    [Component(
        Name = "Sentence Parser",
        Description = "Splits text by sentences (separated by periods)",
        Author = "Satrio Husodo, Jared Payne",
        Version = "1.0",
        ConfigurationType = typeof(SentenceParserConfiguration)
    )]
    [IOSpec(IOSpecType.Output, "outputName", typeof(TLArtifactsCollection))]
    public class SentenceParserComponent : BaseComponent
    {
        private readonly IFileSystem fileSystem;

        public SentenceParserComponent(ComponentLogger logger) : this(logger, new FileSystem()) { }

        internal SentenceParserComponent(ComponentLogger logger, IFileSystem fileSystem) : base(logger)
        {
            this.fileSystem = fileSystem;
            this.Configuration = new SentenceParserConfiguration();
        }

        public new SentenceParserConfiguration Configuration
        {
            get => base.Configuration as SentenceParserConfiguration;
            set => base.Configuration = value;
        }

        public override void Compute()
        {
            var inputFile = this.Configuration.Artifacts.Absolute;
            if (!this.fileSystem.File.Exists(inputFile))
            {
                throw new ComponentException("File path does not exist.");
            }

            var outputDirectory = this.Configuration.OutputDirectory.Absolute;
            if (!this.fileSystem.Directory.Exists(outputDirectory))
            {
                throw new ComponentException("Directory path does not exist.");
            }

            var collection = new TLArtifactsCollection();
            this.fileSystem.File.ReadAllText(inputFile).Split('.')
                .Where(sentence => !string.IsNullOrWhiteSpace(sentence))
                .Select(sentence => sentence.Trim() + ".")
                .Select((sentence, i) => new TLArtifact((i + 1).ToString(), sentence)).ToList()
                .ForEach(collection.Add);
            base.Workspace.Store("outputName", collection);

            collection.Values.Select(artifact => string.Format("S{0}.txt", artifact.Id))
                .Select(filename => this.fileSystem.Path.Combine(outputDirectory, filename))
                .Zip(collection.Values, (path, artifact) => new { Path = path, artifact.Text }).ToList()
                .ForEach(tuple => this.fileSystem.File.WriteAllText(tuple.Path, tuple.Text));
        }

        public IEnumerable<string> GetItems(IEnumerable<string> lines)
        {
            /*
             * Given a list of strings:
             * For each string, put substring in between periods to be in a new list of strings
             */
            var items = new List<string> { "" };
            foreach (var line in lines)
            {
                foreach (var character in line)
                {
                    if (character == '.')
                    {
                        items.Add("");
                    }
                    else
                    {
                        items[items.Count - 1] = items[items.Count - 1] + character;
                    }
                }
            }
            return items;
        }
    }
}
