﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.IO.Abstractions;
using System.Linq;
using TraceLabSDK;
using TraceLabSDK.Types;

namespace SimilarityMatrixVotingMerge
{
    [Component(
        Name = "Similarity Matrix Voting Merge",
        Description = "Merges of a directory of similarity matrices based on a voting percent threshold.",
        Author = "Jared Payne",
        Version = "1.0",
        ConfigurationType = typeof(SimilarityMatrixVotingMergeConfiguration)
    )]
    [IOSpec(IOType = IOSpecType.Output, Name = "MergedMatrix", DataType = typeof(TLSimilarityMatrix))]
    [Tag("Tracers")]
    public class SimilarityMatrixVotingMergeComponent : BaseComponent
    {
        public SimilarityMatrixVotingMergeComponent(ComponentLogger logger) : this(logger, new FileSystem()) { }

        internal SimilarityMatrixVotingMergeComponent(ComponentLogger logger, IFileSystem fileSystem) : base(logger)
        {
            this.Configuration = new SimilarityMatrixVotingMergeConfiguration(fileSystem);
        }

        public new SimilarityMatrixVotingMergeConfiguration Configuration
        {
            get => base.Configuration as SimilarityMatrixVotingMergeConfiguration;
            set => base.Configuration = value;
        }

        public override void Compute()
        {
            var path = this.Configuration.DirectoryPath.Absolute;
            var files = this.Configuration.FileSystem.Directory.GetFiles(path);
            var matrices = files.Select(this.ImportSimilarityMatrix);
            var mergedMatrix = this.Merge(matrices);
            base.Workspace.Store("MergedMatrix", mergedMatrix);
        }

        private TLSimilarityMatrix ImportSimilarityMatrix(string path)
        {
            var lines = this.Configuration.FileSystem.File.ReadAllLines(path);
            var splits = lines.Select(line => line.Split());
            var links = splits.Select(split => new
            {
                SourceId = split[0],
                TargetId = split[1],
                Score = Convert.ToDouble(split[2])
            }).ToList();

            var matrix = new TLSimilarityMatrix();
            links.ForEach(link => matrix.AddLink(link.SourceId, link.TargetId, link.Score));
            return matrix;
        }

        private TLSimilarityMatrix Merge(IEnumerable<TLSimilarityMatrix> matrices)
        {
            var minimumVotesRequiredToIncludeLink = (int)Math.Ceiling(this.Configuration.Threshold * matrices.Count());

            var links = matrices.SelectMany(matrix => matrix.AllLinks);
            var linkGroupings = links.GroupBy(link => new { link.SourceArtifactId, link.TargetArtifactId });
            var filtered = linkGroupings.Where(grouping => grouping.Count() >= minimumVotesRequiredToIncludeLink);
            var averages = filtered.ToDictionary(grouping => grouping.Key, grouping => grouping.Select(link => link.Score).Average()).ToList();

            var result = new TLSimilarityMatrix();
            averages.ForEach(pair => result.AddLink(pair.Key.SourceArtifactId, pair.Key.TargetArtifactId, pair.Value));
            return result;
        }
    }
}
