# How to write unit tests for TraceLab components written in C#.

1. In Visual Studio, open the TraceLab component's solution.
2. Add a new Unit Test Project (.NET Framework) to the solution.
3. In the generated unit test project, add references to the original component
   project and the TraceLabSDK DLL file.
4. Using NuGet, install the package `Moq` to the unit test project.
5. In the provided source file, add the following code:

    ```csharp
    using Microsoft.VisualStudio.TestTools.UnitTesting;
    using Moq;
    using System.Collections.Generic;
    using TraceLabSDK;
    using TraceLabSDK.Types;

    namespace MyComponent.Tests
    {
        [TestClass]
        public class MyComponentTests
        {
            private IDictionary<string, object> workspace;

            private MyComponent component;

            [TestInitialize]
            public void TestInitialize()
            {
                this.workspace = new Dictionary<string, object>();

                // Creates a component using a mock logger.
                var logger = new Mock<ComponentLogger>();
                this.component = new MyComponent(logger.Object);

                // Initializes the workspace with a mock that just uses a dictionary.
                var workspace = new Mock<IWorkspace>();
                workspace.Setup(mock => mock.Load(It.IsAny<string>()))
                    .Returns<string>(str => this.workspace[str]);
                workspace.Setup(mock => mock.Store(It.IsAny<string>(), It.IsAny<object>()))
                    .Callback<string, object>((str, obj) => this.workspace[str] = obj);
                this.component.Workspace = workspace.Object;
            }

            // This is just an example test. It works!
            [TestMethod]
            public void TestWorkspace()
            {
                this.component.Workspace.Store("Hello", "World!");
                Assert.AreEqual("World!", this.component.Workspace.Load("Hello"));
            }
        }
    }
    ```

    This code creates a component with a dummy logger and simulated workspace.

## How to mock the filesystem

1. Using NuGet, install the package `System.IO.Abstractions` to both
   projects and `System.IO.Abstractions.TestingHelpers` to the test project.
2. Add the following code to your component class:

   ```csharp
   using System.IO.Abstractions;

   namespace MyComponent
   {
       public class MyComponent : BaseComponent
       {
           private readonly IFileSystem fileSystem;

           // The normal constructor used by TraceLab.
           public MyComponent(ComponentLogger logger) : this(logger, new FileSystem()) { }

           // The filesystem injection constructor used for testing.
           internal MyComponent(ComponentLogger logger, IFileSystem fileSystem) : base(logger)
           {
               this.fileSystem = fileSystem;
           }
       }
   }
   ```

   Wherever you would normally make a call to a `System.IO` method (e.g.,
   `Directory.Exists`, `File.ReadLines`), prefix it with `this.fileSystem`
   to place the call to the proper filesystem.

3. In your component project's `AssemblyInfo.cs` file, add the line:

   ```csharp
   [assembly: InternalsVisiableTo("MyComponent.Tests")]
   ```

3. In the unit test class, a mock filesystem can be implemented:

    ```csharp
    using Microsoft.VisualStudio.TestTools.UnitTesting;
    using Moq;
    using System.IO.Abstractions.TestingHelpers;
    using TraceLabSDK;

    namespace MyComponent.Tests
    {
        [TestClass]
        public class MyComponentTests
        {
            private MockFileSystem fileSystem;

            private MyComponent component;

            [TestInitialize]
            public void TestInitialize()
            {
                this.fileSystem = new MockFileSystem();

                var logger = new Mock<ComponentLogger>();
                this.component = new MyComponent(logger.Object, this.fileSystem);
            }

            [TestMethod]
            public void TestMockFileSystem()
            {
                var mockFileData = new MockFileData("File contents here.");
                this.fileSystem.AddFile("file.txt", mockFileData);
            }
        }
    }
    ```

## Using NuGet packages with TraceLab components

NuGet packages are always dynamically linked to .NET projects. Therefore, if
part of a NuGet package os used in a TraceLab component project, the component
may fail to work in TraceLab because the DLL will not be found.

In order to circumvent this issue, a potential workaround is to import the
source code of the NuGet package into its own library project within your
solution. A reference can then be added from the dependent project to the
imported library and the library can be utilized as normal. Finally, make sure
that the library project's target DLL file is copied to the TraceLab component
directory as a post-build step, much like any other TraceLab component DLL file.
